#include "ehs/io/hid/HID.h"

namespace ehs
{
	HID::HID()
		: type(EHS_HID_UNKNOWN), id(0)
	{
	}

	HID::HID(const UInt_8 type, Str_8 name, const UInt_64 id)
		: type(type), name((Str_8&&)name), id(id)
	{
	}

	HID::HID(HID&& hid) noexcept
		: type(hid.type), name((Str_8&&)hid.name), id(hid.id), states((Array<ButtonState>&&)hid.states)
	{
		hid.type = EHS_HID_UNKNOWN;
		hid.id = 0;
	}

	HID::HID(const HID& hid)
		: type(hid.type), name(hid.name), id(hid.id), states(hid.states)
	{
	}

	HID& HID::operator=(HID&& hid) noexcept
	{
		if (this == &hid)
			return *this;

		type = hid.type;
		name = (Str_8&&)hid.name;
		id = hid.id;
		states = (Array<ButtonState>&&)hid.states;

		hid.type = EHS_HID_UNKNOWN;
		hid.id = 0;

		return *this;
	}

	HID& HID::operator=(const HID& hid)
	{
		if (this == &hid)
			return *this;

		type = hid.type;
		name = hid.name;
		id = hid.id;
		states = hid.states;

		return *this;
	}

	bool HID::operator==(const HID& other) const
	{
		return type == other.type && hashName == other.hashName;
	}

	bool HID::operator!=(const HID& other) const
	{
		return type != other.type || hashName != other.hashName;
	}

	bool HID::operator==(const UInt_64 otherId) const
	{
		return id == otherId;
	}

	bool HID::operator!=(const UInt_64 otherId) const
	{
		return id != otherId;
	}

	void HID::Poll()
	{
		for (UInt_64 i = 0; i < states.Size(); ++i)
		{
			if (states[i].IsPressed())
			{
				if (states[i].GetState() == State::RELEASED)
					states[i].SetState(State::TOUCHED);
				else
					states[i].SetState(State::PRESSED);
			}
			else
			{
				if (states[i].GetState() == State::PRESSED || states[i].GetState() == State::TOUCHED)
					states[i].SetState(State::JUST_RELEASED);
				else
					states[i].SetState(State::RELEASED);
			}
		}
	}

	UInt_8 HID::GetType() const
	{
		return type;
	}

	Str_8 HID::GetName() const
	{
		return name;
	}

	UInt_64 HID::GetId() const
	{
		return id;
	}

	void HID::ReleaseAll()
	{
		for (UInt_64 i = 0; i < states.Size(); ++i)
			states[i].SetPressed(false);
	}

	Vector<const ButtonState*> HID::GetAllTouched() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::TOUCHED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsTouched(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || *state != State::TOUCHED)
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsTouched() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::TOUCHED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllDown() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::PRESSED || states[i] == State::TOUCHED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsDown(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || (*state != State::PRESSED && *state != State::TOUCHED))
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsDown() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::PRESSED || states[i] == State::TOUCHED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllJustReleased() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::JUST_RELEASED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsJustReleased(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || *state != State::JUST_RELEASED)
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsJustReleased() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::JUST_RELEASED)
				return &states[i];

		return nullptr;
	}

	Vector<const ButtonState*> HID::GetAllUp() const
	{
		Vector<const ButtonState*> result(0, states.Size() + 1);

		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::RELEASED || states[i] == State::JUST_RELEASED)
				result.Push(&states[i]);

		return result;
	}

	const ButtonState* HID::IsUp(const Button& button) const
	{
		const ButtonState* state = GetState(button);
		if (!state || (*state != State::RELEASED && *state != State::JUST_RELEASED))
			return nullptr;

		return state;
	}

	const ButtonState* HID::IsUp() const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i] == State::RELEASED || states[i] == State::JUST_RELEASED)
				return &states[i];

		return nullptr;
	}

	void HID::ButtonDown(const Button& button)
	{
		if (ButtonState* state = GetState(button); state)
			state->SetPressed(true);
		else
			states.Push(ButtonState(button, State::RELEASED));
	}

	void HID::ButtonUp(const Button& button)
	{
		if (ButtonState* state = GetState(button); state)
			state->SetPressed(false);
		else
			states.Push(ButtonState(button, State::JUST_RELEASED));
	}

	const ButtonState* HID::GetState(const Button& button) const
	{
		for (UInt_64 i = 0; i < states.Size(); ++i)
			if (states[i] == button)
				return &states[i];

		return nullptr;
	}

	bool HID::IsValid() const
	{
		return id;
	}

	HID* HID::Clone() const
	{
		return new HID(*this);
	}

	bool HID::HasState(const Button& button) const
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i].GetButton() == button)
				return true;

		return false;
	}

	bool HID::AddState(const ButtonState& state)
	{
		if (HasState(state.GetButton()))
			return false;

		states.Push(state);

		return true;
	}

	ButtonState* HID::GetState(const Button& button)
	{
		for (UInt_64 i = 0; i < states.Size(); i++)
			if (states[i].GetButton() == button)
				return &states[i];

		return nullptr;
	}
}